@@ -1,32 +1,29 @@
 import asyncio
-from enum import Enum
 import aiohttp
 from dataclasses import dataclass
 import ipaddress
 import logging
 import platformdirs
 import random
-import re
 import socket
-import sys
 import os
-from typing import Optional
+from sqlmodel import Session, SQLModel, create_engine, select
 import warnings
-from sqlmodel import Field, Session, SQLModel, create_engine, select
 from grep_ips.grep_ips import GrepIPs
+from .parrots import Parrot
 
 
-def cli():
+def cli(return_help_text=False):
     import argparse  # https://docs.python.org/3/library/argparse.html
 
-    help_width = None if sys.stdout.isatty() else 78  # consistent width for README.py
+    help_width = 78 if return_help_text else None  # consistent width for README.py
     formatter_class = lambda prog: argparse.HelpFormatter(
         prog,
         max_help_position=37,
         width=help_width,
     )
     parser = argparse.ArgumentParser(
-        description="Fast, fault-tolerant public IP address retrieval from Python or CLI.",
+        prog=app_name(),
         formatter_class=formatter_class,
     )
     parser.add_argument(
@@ -60,6 +57,13 @@ def cli():
         default=10,
         help="maximum number of simultaneous network connections allowed (default: 10)",
     )
+    db_file_display = db_pathname().replace(os.path.expanduser('~'), '~')
+    parser.add_argument(
+        "--dbfile",
+        type=str,
+        default='',  # need to call db_pathname() again later with create_dir=True
+        help=f"path for database file ('-' for memory-only; default: {db_file_display})",
+    )
     parser.add_argument(
         "--show-api-providers",
         action='store_true',
@@ -102,9 +106,12 @@ def cli():
         const=1,
         help="increase verbosity",
     )
+    if return_help_text:  # used by README.py
+        return parser.format_help()
     args = parser.parse_args()
     if args.show_api_providers:
-        Parrot.show_parrot_db()
+        engine = init_db(args.dbfile)
+        Parrot.show_parrot_db(engine)
         return
     del args.show_api_providers
     if args.ipv4:
@@ -120,20 +127,34 @@ def cli():
 
 
 @dataclass
-class MyoriginArgs:
+class MyoriginArgs:  # see '--help' for descriptions
     timeout: int = 12000
     minimum_match: int = 2
     overkill: int = 0
     max_failures: int = 10
     max_connections: int = 10
+    dbfile: str = ''  # ''==db_pathname(), '-'==none, other==alternate file
     ip_version: int = 0  # 0==either, 4==IPv4 only, 6==IPv6 only
     logfile: str = '-'
     log_level: int = 0  # 0==disabled, 1==errors, 2==warnings, 3==info, 4==debug
+    exception_level = 0  # 0==disabled, 1==raise exception for errors, 2==... for warnings, etc.
+
+
+def app_name():
+    return os.path.splitext(os.path.basename(__file__))[0]
+
+
+class NetworkError(Exception):
+    pass
 
 
 def my_ip(args: MyoriginArgs) -> str:
+    if args.log_level >= 3:  # info or debug
+        log_format = '%(asctime)s.%(msecs)03d %(levelname)s %(message)s'
+    else:
+        log_format = '%(message)s'
     logging.basicConfig(
-        format='%(asctime)s.%(msecs)03d %(levelname)s %(message)s',
+        format=log_format,
         datefmt='%H:%M:%S',
         filename=args.logfile if args.logfile != '-' else None,
         filemode='a',
@@ -144,7 +165,10 @@ def my_ip(args: MyoriginArgs) -> str:
         logger.setLevel(log_levels[args.log_level])
     except IndexError:
         logger.setLevel(logging.WARNING)
-        logger.error(f"Invalid log level")
+        error = "Invalid log level"
+        logger.error(error)
+        if args.exception_level >= 1:
+            raise ValueError(error)
         return ""
     return asyncio.run(main_loop(args, logger))
 
@@ -201,102 +225,6 @@ async def http_get(url: str, session: aiohttp.ClientSession, max_size=-1, ip_ver
         raise ValueError(s if len(s) > 1 else "Unknown exception in http_get()")
 
 
-class Parrot(SQLModel, table=True):  # data for one interface of an API provider
-    # (Parrots repeats what they hear. A my-IP API provider repeats the callers IP back to them.)
-    id: int = Field(primary_key=True)
-    ptask: str  # one of: 'I', 'J', 'D', 'S', 'M', 'N'
-    ip_version: int  # always 4 or 6
-    address: str  # host+path (URL without protocol)
-    milliweight: int = 1000  # 2000 means 2x more likely to be used; 0 means disabled
-    attempt_count: int = 0  # number of attempted connections
-    success_count: int = 0  # number of valid IP addresses returned
-    total_rtt: int = 0  # total rtt (in ms) for all successful attempts
-    last_errmsg: str = ""
-
-    @staticmethod
-    def startup():
-        with Session(engine) as session:
-            for line in parrot_data.split('\n'):
-                wout_comments = re.sub(r'( +|^)#.*\n?', '', line)  # strip comments
-                if len(wout_comments) == 0:
-                    continue
-                fields = re.sub(r' +', ' ', wout_comments).split(' ')
-                assert len(fields) >= 3, f"invalid parrot_data line: {line}"
-                # do not adjust the Ptask list here except to append; .id is computed from it
-                for i, ptask in enumerate([Ptask.IP_HTTP, Ptask.IP_HTTPS]):
-                    address = fields[1]
-                    ip_version = int(fields[2])
-                    for v in (4, 6):  # IPv4, IPv6
-                        ip_version_offset = 0 if v == 4 else 50  # IPv4 uses 0-49, IPv6 uses 50-99
-                        id = 2018260000 + int(fields[0]) * 100 + ip_version_offset + i
-                        statement = select(Parrot).where(Parrot.id == id)
-                        result = session.exec(statement).one_or_none()
-                        ptask_active = ptask.value in fields[3:]
-                        ip_version_active = ip_version == 0 or ip_version == v
-                        if result is None:  # no existing row in DB
-                            if not (ptask_active and ip_version_active):
-                                continue  # not in DB and shouldn't be
-                            result = Parrot()  # add new row to database for this Ptask
-                            result.id = id
-                        else:  # existing row in DB
-                            if not (ptask_active and ip_version_active):
-                                result.milliweight = 0  # disable in DB
-                            else:
-                                result.milliweight = 1000  # make sure it is enabled
-                        result.ptask = ptask.value
-                        result.ip_version = v
-                        result.address = address
-                        session.add(result)
-            session.commit()
-
-    def url(self):
-        if self.ptask == Ptask.IP_HTTP.value:
-            return 'http://' + self.address
-        if self.ptask == Ptask.IP_HTTPS.value:
-            return 'https://' + self.address
-        assert ValueError, f"Ptask {Ptask(self.ptask).name} not yet implemented"
-
-    def score(self):  # compute a score which will determine how likely it is to be chosen
-        if self.attempt_count != 0:
-            percent = 100.0 * self.success_count / self.attempt_count
-        else:
-            percent = 100.0
-        score = round(percent**2 / 100.0)  # biggest portion of score is success rate
-        score += 5  # every parrot gets a small chance of being selected
-        if self.attempt_count < 10:
-            score += 5  # prefer new parrots
-        if self.success_count != 0:
-            average_ms = 1.0 * self.total_rtt / self.success_count
-            score += round(max((5000 - average_ms) / 400, 0.0))  # prefer faster parrots
-        score *= self.milliweight  # normally 1000, but can be 0 to disable or more to promote
-        return score
-
-    @staticmethod
-    def show_parrot_db():
-        with Session(engine) as session:
-            to_show = dict()
-            results = session.exec(select(Parrot))
-            for p in results:
-                score = p.score()
-                disp = ""
-                disp += f"ipv{p.ip_version}.{p.url()}:\n"
-                if p.attempt_count != 0:
-                    percent = round(100.0 * p.success_count / p.attempt_count)
-                    disp += f"    {percent}% success rate"
-                    disp += f" ({p.success_count} of {p.attempt_count})\n"
-                else:
-                    disp += f"    not yet attempted\n"
-                if p.success_count != 0:
-                    average_ms = round(1.0 * p.total_rtt / p.success_count)
-                    disp += f"    {average_ms} ms average round trip time\n"
-                if len(p.last_errmsg) > 0:
-                    disp += f"    most recent error: {p.last_errmsg[:70]}\n"
-                disp += f"    score: {round(score/1000):,} points\n"
-                to_show[score * 1000000 + p.id] = disp  # sortable, unique key
-            for p in dict(sorted(to_show.items(), reverse=True)):
-                print(to_show[p], end="")
-
-
 @dataclass
 class FetchedIP:
     parrot: Parrot
@@ -349,84 +277,6 @@ async def get_ip(
     await q.put(fip)
 
 
-class Ptask(Enum):
-    IP_HTTP = 'I'  # get IP via http
-    IP_HTTPS = 'J'  # get IP via https
-    IP_DNS = 'D'  # get IP via DNS (not yet implemented)
-    IP_STUN = 'S'  # get IP via STUN (not yet implemented)
-    MD_HTTP = 'M'  # get IP metadata via http (not yet implemented)
-    MD_HTTPS = 'N'  # get IP metadata via https (not yet implemented)
-
-
-parrot_data = '''
-# **Do not change or reuse IDs** (fields[0])
-
-# fields[0] → parrot number (always unique so we can safely copy future changes to user's DB)
-# fields[1] → parrot address (URL without 'https://' or 'http://')
-# fields[2] → ip version (0==both, 4==IPv4 only, 6==IPv6 only)
-# fields[3:] → Ptask list
-
-# currently, each row can expand to 4 database rows (2 IP versions × http vs. https)
-
-00 icanhazip.com/                            0 I J
-01 ip1.dynupdate.no-ip.com/                  4 I
-02 myip.dnsomatic.com/                       4 I J
-03 smart-ip.net/myip                         0             # cannot connect
-04 ipecho.net/plain                          4   J
-05 ident.me/                                 0 I J M N     # https://api.ident.me/
-06 tnedi.me/                                 0 I J M N     # https://ipa.tnedi.me/
-07 ip.appspot.com/                           0             # 503 Service Unavailable
-08 checkip.dyndns.org/                       4 I
-09 www.lawrencegoetz.com/programs/ipinfo/    0             # no API
-10 shtuff.it/myip/short/                     0             # cannot connect
-11 ifconfig.me/ip                            4 I J
-12 www.google.com/search?q=my+ip             0             # 403 Forbidden
-13 bot.whatismyipaddress.com/                0             # API discontinued "due to massive abuse"
-14 ipv4.ipogre.com/                          0             # Connection timeout
-15 automation.whatismyip.com/n09230945.asp   0             # API discontinued?
-16 myipis.net/                               0             # API discontinued?
-17 www.ipchicken.com/                        4   J
-18 myip.com.tw/                              4   J
-19 httpbin.org/ip                            4 I J
-20 ip.nf/me.txt                              4   J   N
-21 am.i.mullvad.net/ip                       0   J
-22 am.i.mullvad.net/json                     0       N     # https://mullvad.net/en/check/
-23 zx2c4.com/ip                              0 I J
-24 ip.websupport.sk/                         0 I J
-25 www.ivpn.net/                             0   J
-26 www.ipaddress.com/                        0             # API discontinued?
-27 www.ipaddress.my/                         0   J
-28 www.showmyip.com/                         0             # no API
-29 ip-api.com/line/?fields=query             4 I
-30 ip-api.com/json/?fields=16966359          0     M       # https://ip-api.com/docs/
-31 api.ipify.org/                            4 I J
-32 ifconfig.io/ip                            0 I J
-33 ipaddress.sh/                             4 I J
-34 api.iplocation.net/?ip={ip}               0       N     # https://api.iplocation.net/
-35 ipinfo.io/ip                              4 I J
-36 ipinfo.io/{ip}/json                       4       N
-37 api.ipregistry.co/?key=tryout             0 I J   N     # https://ipregistry.co/docs/
-38 myexternalip.com/raw                      4 I J
-39 checkip.amazonaws.com/                    4 I J
-40 diagnostic.opendns.com/myip               0             # cannot connect
-41 whatismyip.akamai.com/                    0             # cannot connect
-42 test-ipv6.com/ip/                         4 I J         # ironically, no IPv6 AAAA record
-43 api.infoip.io/                            4 I J M N     # https://ciokan.docs.apiary.io/
-44 checkip.dns.he.net/                       0 I J
-45 ipapi.co/ip                               0   J
-46 www.cloudflare.com/cdn-cgi/trace          0 I J
-47 www.trackip.net/ip                        0 I J
-48 www.trackip.net/ip?json                   0     M N
-49 mypubip.com/                              0 I J
-50 ip.seeip.org/                             0   J
-51 ip.seeip.org/geoip                        0     M N
-52 api.bigdatacloud.net/data/client-ip       4 I J
-53 myip.opendns.com%20A%20@resolver1.opendns.com   0   D  
-54 o-o.myaddr.l.google.com%20TXT%20@8.8.8.8        0   D  
-55 whoami.akamai.net%20ANY%20@ns1-1.akamaitech.net 0   D  
-'''
-
-
 def weighted_random(options: dict):
     # return a randomly-chosen key with the probability represented by its value
     total = sum(options.values())
@@ -449,6 +299,8 @@ class DoneWithJobs(Exception):
 
 async def main_loop(args: MyoriginArgs, logger: logging.Logger) -> str:
     result = ""
+    error = None
+    engine = init_db(args.dbfile)
     with Session(engine) as db_session:
         scores = dict()
         if args.ip_version == 0:
@@ -490,7 +342,7 @@ async def main_loop(args: MyoriginArgs, logger: logging.Logger) -> str:
                         max_ip_count = 0 if len(ip_counts[v]) == 0 else max(ip_counts[v].values())
                         # FIXME: for multiple IPs, maybe ensure most_common >= second_most_common + minimum_match more
                         if len(ip_counts[v]) > 1:
-                            logger.error(f"multiple IPs received: {ip_counts[v]}")
+                            error = f"multiple IPs received: {ip_counts[v]}"
                             raise DoneWithJobs
                         if max_ip_count >= args.minimum_match:  # we have sufficient results
                             ip = max(ip_counts[v], key=ip_counts[v].get)
@@ -515,17 +367,17 @@ async def main_loop(args: MyoriginArgs, logger: logging.Logger) -> str:
                         # TCPConnector(limit=...) does this too, but checking args.max_connections
                         # ... here delays "not enough parrots" so we can collect more responses
                         if len(scores) == 0:
-                            logger.error(f"not enough parrots for {args.minimum_match} matches")
+                            error = f"not enough parrots for {args.minimum_match} matches"
                             raise DoneWithJobs
                         p_id = weighted_random(scores)
                         statement = select(Parrot).where(Parrot.id == p_id)
                         p = db_session.exec(statement).one_or_none()
                         assert p is not None
-                        a = p_id - (p_id % 100)  # 100 IDs per parrot_data line
-                        b = a + 100  # top of range for a sinle parrot_data line
+                        a = p_id - (p_id % 100)  # 100 IDs per flock_data line
+                        b = a + 100  # top of range for a sinle flock_data line
                         ids_to_del = [k for k in scores.keys() if a <= k < b]
                         logger.debug(f"chose {p_id}, removing {ids_to_del}")
-                        for k in ids_to_del:  # eliminate all IDs from same parrot_data line, e.g.
+                        for k in ids_to_del:  # eliminate all IDs from same flock_data line, e.g.
                             del scores[k]  # don't check both http://ident.me and https://ident.me
                         logger.debug(f"launching task IPv{p.ip_version} {p.url()}")
                         asyncio.create_task(get_ip(p, aio_session, q, ip_version=args.ip_version))
@@ -554,7 +406,7 @@ async def main_loop(args: MyoriginArgs, logger: logging.Logger) -> str:
                         db_session.add(r.parrot)
                         db_session.commit()
                         if fail_count >= args.max_failures:
-                            logger.error(f"{fail_count} requests failed; giving up")
+                            error = f"{fail_count} requests failed; giving up"
                             raise DoneWithJobs
                     except asyncio.QueueEmpty:
                         await asyncio.sleep(0.005)  # reduce looping when waiting on responses
@@ -562,33 +414,41 @@ async def main_loop(args: MyoriginArgs, logger: logging.Logger) -> str:
             except DoneWithJobs:
                 pass
         connector.close()
+    if error is not None:
+        logger.error(error)
+        if args.exception_level >= 1:
+            raise NetworkError(error)
     return result
 
 
-engine = None
+def db_pathname(create_dir=False):
+    config_dir = platformdirs.user_config_dir(app_name())
+    if create_dir:
+        try:
+            os.mkdir(config_dir)
+        except FileExistsError:
+            pass
+        # except (PermissionError, FileNotFoundError):
+        #     one of these will be raised if the directory cannot be created
+    return os.path.join(config_dir, f'data.sqlite')
 
 
-def on_import():
-    global engine
-    appname = 'myorigin'
+def init_db(db_file=''):
+    if hasattr(init_db, 'engine'):  # only need to initialize once
+        return init_db.engine
     GrepIPs.grep_ips_test()
     warnings.filterwarnings(  # https://github.com/tiangolo/sqlmodel/issues/189#issuecomment-1018014753
         "ignore", ".*Class SelectOfScalar will not make use of SQL compilation caching.*"
     )
-    config_dir = platformdirs.user_config_dir(appname)
-    try:
-        os.mkdir(config_dir)
-    except FileExistsError:
-        pass
-    # except (PermissionError, FileNotFoundError):
-    #     one of these will be raised if the directory cannot be created
-    db_file = os.path.join(config_dir, f'data.sqlite')
-    engine = create_engine(f'sqlite:///{db_file}', echo=False)
-    SQLModel.metadata.create_all(engine)
-    Parrot.startup()
+    if db_file == '':  # use default location
+        db_file = db_pathname(create_dir=True)
+    elif db_file == '-':
+        db_file = ':memory:'
+    init_db.engine = create_engine(f'sqlite:///{db_file}', echo=False)
+    SQLModel.metadata.create_all(init_db.engine)
+    Parrot.startup(init_db.engine)
+    return init_db.engine
 
 
-on_import()
-
 if __name__ == "__main__":
     cli()
